/*
 API | rdxSort.h | Fonctions de tris avec la methode Radix Sort
 Editeur : Ajanohoun Jordy 
 Ecole : Polytech Sorbonne 
 Cours : Algorithme General by Cecile Braunstein  */

# ifndef RDX_SORT_H
# define RDX_SORT_H

#include "queue.h"
#include <math.h>


/*  Calcul et retourne le nombre de chiffres du plus grand nombre en valeur absolue du tableau d'entiers Tab passe en argument avec sa taille.
Le tableau n'est pas modifie */
int getTailleClefInt(int *Tab, const int SIZE);



/*  Implementation du Radix Sort LSD pour trier un tableau d'entiers POSITIFS (Tab) dans l'ordre croissant. Le tri s'effectue au niveau binaire, c'est a dire que les iterations pour trier se font sur un bit a la fois en allant du bit de poids le plus faible au bit de poids le plus fort.
- SIZE est la taille du tableau
- TAILLE_CLEF est le nombre de chiffre du plus grand nombre en valeur absolue du tableau. Si cette valeur n'est pas connue alors passer un chiffre <= 0 a la place en argument et la fonction se chargera de la calculer avant d'effectuer le tri radix sort. 
-  Le tableau est modifier, il est trie a la sortie de la fonction 
- Utilise un autre tableau 2 fois plus grand que Tab pour stocker les donnes intermediaire a chaque iteration  */
void rdxSortPosIntTab_Bi(int *Tab, const int SIZE, int TAILLE_CLEF);




/*  Meme chose que rdxSortIntTabBi a un detail pret : 
- utilise une structure de donnee file (FIFO) pour stocker le "tableau" intermediaire (les resultats de l'etape intermediaire) au lieu d'un autre tableau 2 fois plus grand que l'initial   */
void rdxSortPosIntFile_Bi (int *Tab, const int SIZE, int TAILLE_CLEF);



/*  Implementation du Radix Sort en base 10 cette fois ci. Pour trier tableau de nombres POSITIFS
- les iterations pour trier les nombres se font sur un digit (dans la base 10) a la fois en commencant par le chiffre des unités.
- SIZE est la taille du tableau
- TAILLE_CLEF est le nombre de chiffre du plus grand nombre en valeur absolue du tableau. Si cette valeur n'est pas connue alors passer un chiffre <= 0 a la place en argument et la fonction se chargera de la calculer avant d'effectuer le tri
-  Le tableau est modifier, il est trie a la sortie de la fonction 
- Utilise une file (FIFO) pour stocker les donnees a chaque iteration   */
void rdxSortPosIntFile_De(int *Tab, const int SIZE, int TAILLE_CLEF);



/*  Radix Sort en base 10 exactement comme la fonction rdxSortPosIntFile_De MAIS pour les nombres negatifs. Tab doit etre un tableau de nombres < 0 uniquement  */
void rdxSortNegIntFile_De(int *Tab, const int SIZE, int TAILLE_CLEF);



/*
  Radix Sort pour un tableau d'entiers positifs ET/OU negatifs
  - utilise toutes les fonctions de radix sort specifique precedentes
  - tri les nombres negatifs et positifs separement puis fusionne le tout pour avoir le tableau d'origine trie entierement
  -rdxSortType est compris entre 0 et 2 et permet de choisir avec quelle fonction specifique de radix sort trier la partie des nombres positifs 
     0 pour rdxSortPosIntTab_Bi | 1 pour rdxSortPosIntFile_Bi | 2 pour rdxSortPosIntFile_De
  - La partie negative est toujours triee avec la fonction rdxSortNegIntFile_De
  */
void rdxSortIntChoix(int *Tab, const int SIZE, int TAILLE_CLEF, int rdxSortType);


# endif
