/*
 Code source | rdxSort.c | Fonctions de tris avec la methode Radix Sort
 Editeur : Ajanohoun Jordy 
 Ecole : Polytech Sorbonne 
 Cours : Algorithme General by Cecile Braunstein  */

#include "rdxSort.h"

int getTailleClefInt(int *Tab, const int SIZE){
  int i, res=1, temp, diviseur;

  // Parcours du tableau 
  for(i=0; i<SIZE; i++){
    temp = 1; // variable temporaire qui pour chaque nombre du tableau va stocker son nombre de chiffre 
    diviseur = 10;

    // Calcul du nombre de digits composant le nombre dans la case d'indice i 
    while(((int)(Tab[i]/diviseur)) != 0 ){
      temp++;
      diviseur *= 10;
    }

    if(temp > res){
      res = temp; // on stock dans le resultat le plus grand nombre de chiffres composant un nombre parmi ceux du tableau 
    }
  }
  
  return res;
}



void rdxSortPosIntTab_Bi (int *Tab, const int SIZE, int TAILLE_CLEF){

  if(TAILLE_CLEF <= 0){ // si la taille maximum des clefs n'est pas connue on la calcul avant d'effectuer le tri
    TAILLE_CLEF = getTailleClefInt(Tab, SIZE);
  }
  
  // calcul du nombre de bits a traiter pour trier correctement l'ensemble du tableau
  // depend de TAILLE_CLEF
  // cf rdxSort.h et le rapport pour plus d'informations sur ce calcul
  int nbBits =(TAILLE_CLEF +1.0)/ log10(2) +1;

  
  int iter, i, j, i2;
  
  int *tabTemp = (int *)malloc(2*SIZE*sizeof(int)); // tableau 2 fois plus grand 

  // Traitement bit a bit  
  for(iter = 1; iter <= nbBits; iter++){
    i = 0; j = 0;

    while(i+j < SIZE){

      // on recupere le bit a la position iter en partant du bit de poids le plus faible pour chacun des nombres du tableau 
      
      if ((Tab[0]% (int)(pow(2,iter))) >> (iter-1)){
	tabTemp[i]=Tab[0]; // s'il est egal a 1 on copie le nombre correspondant dans la moitier gauche du tableau tabtemp
	i++; // indice qui nous indique combien il y a d'elements dans la moitier gauche du tableau tabTemp
      }
      else{
	tabTemp[SIZE+j]=Tab[0]; // s'il est egal a 0 on fait de meme dans la moitie droite de tabTemp
	j++; // indice qui indique combien d'elemnts il y a effectivement dans la moitier droite de tabTemp
      }
      
      Tab++; // deplacement au nombre suivant dans le tableau a trier
      
    }
    
    Tab = Tab - SIZE; // on revient au debut du tableau pour traiter le bit suivant (bit a la position iter+1)

    
    // on reforme le tableau d'origine mais trie jusqu'au bit iter-ieme bit
    for(i2=0;i2<j;i2++){ // d'abord on y met les nombres dont le bit iter vaut 0 et dans l'ordre dans lesquels on les a rencontre dans le tableau ====> tri stable
      Tab[i2]=tabTemp[SIZE+i2];
    }
    
    for(i2; (i2-j)<i; i2++){ // puis ceux dont le bit iter est egal a 1 et dans l'ordre dans lesquels on les a rencontres egalement puisque on les stock dans tempTab dans cet ordre (cf la boucle while juste au dessus pour s'en rendre compte)
      Tab[i2]=tabTemp[i2-j];
      }
  }
  
  free(tabTemp);
}



void rdxSortPosIntFile_Bi (int *Tab, const int SIZE, int TAILLE_CLEF){

  if(TAILLE_CLEF <= 0){ // si la taille maximum des clefs n'est pas connue on la calcul avant d'effectuer le tri
    TAILLE_CLEF = getTailleClefInt(Tab, SIZE);
  }

  Queue maFile0 = initQueue(); // file pour les bits qui valent 0
  Queue maFile1 = initQueue(); // file pour les bits qui valent 1

  // calcul du nombre de bits a traiter pour trier correctement l'ensemble du tableau
  // depend de TAILLE_CLEF
  // cf rdxSort.h et le rapport pour plus d'informations sur ce calcul
  int nbBits =(TAILLE_CLEF +1.0)/ log10(2) +1;
  
  int iter, i;

  // Traitement bit a bit
  for(iter = 1; iter <= nbBits; iter++){
    i=0;
    while(i< SIZE){
      // on recupere le bit a la position iter en partant du bit de poids le plus faible pour chacun des nombres du tableau
      if ((Tab[i]% (int)(pow(2,iter))) >> (iter-1)){
	enqueue(&maFile1, Tab[i]); // s'il est egal a 1 on enfile le nombre correspondant dans la file correspondante (file des nombres dont le bit iter vaut 1 )  
      }
      else{
	enqueue(&maFile0, Tab[i]); // s'il est egal a 0 on enfile le nombre correspondant dans la file correspondante (file des nombres dont le bit iter vaut 0 )
      }
      i++;
    }

    // on reforme le tableau d'origine mais trié jusqu'au bit iter-ieme bit
    i=0;
      while(maFile0.head){
	Tab[i] = dequeue(&maFile0); // on defile et range d'abord dans le tableau les nombres dont le bit iter vaut 0 et dans l'ordre dans lequel ils ont ete enfilé (tri stable)
	i++;
      }
      while(maFile1.head){
	Tab[i]=dequeue(&maFile1); // on defile et range ensuite les nombres dont le bit iter vaut 1 egalement dans l'ordre (tri stable)
	i++;
      }
  }
}



void rdxSortPosIntFile_De (int *Tab, const int SIZE, int TAILLE_CLEF){
  
  if(TAILLE_CLEF <= 0){// si la taille maximum des clefs n'est pas connue on la calcul avant d'effectuer le tri
    TAILLE_CLEF = getTailleClefInt(Tab, SIZE);
  }
  
  Queue fileTab[10]; // tableau des files correspondantes aux chiffres de 0 a 9 
  int i, j, diviseur, k;
  for(i=0; i<10; i++){
    fileTab[i] = initQueue(); // initialisation des files
  }

  // on repete pour chaque digit 
  for(i=0, diviseur=1; i<TAILLE_CLEF; i++,diviseur*=10){
    
    for(j=0; j<SIZE; j++){
      
      enqueue(&fileTab[((int)(Tab[j]/diviseur))%10], Tab[j]); // on evalue le i-eme chiffre du nombre et on enfile le nombre dans la file correspondante a ce chiffre
    }
    
    // on reforme le tableau d'origine mais trié jusqu'au i-eme chiffre des nombres en partant du chiffre des unités 
    k =0;
    for(j=0; j<10; j++){ // on defile et range dans le tableau tous les nombres des 10 files dans l'ordre croissant des chiffres (d'abord file 0, 1,2,3 ... ) et de maniere a ce que ce soit un tri stable car on utilise la structure FIFO 
      while(fileTab[j].head){
	Tab[k] = dequeue(&fileTab[j]);
	k++;
	
      }
    }
  }
}


void rdxSortNegIntFile_De (int *Tab, const int SIZE, int TAILLE_CLEF){
  
  if(TAILLE_CLEF <= 0){// si la taille maximum des clefs n'est pas connue on la calcul avant d'effectuer le tri
    TAILLE_CLEF = getTailleClefInt(Tab, SIZE);
  }
  
  Queue fileTab[10]; // tableau des files correspondantes aux chiffres de 0 a 9 
  int i, j, diviseur, k;
  for(i=0; i<10; i++){
    fileTab[i] = initQueue(); // initialisation des files
  }

  // on repete pour chaque digit 
  for(i=0, diviseur=1; i<TAILLE_CLEF; i++,diviseur*=10){
    
    for(j=0; j<SIZE; j++){
      
      enqueue( & fileTab[ abs( ((int) (Tab[j]/diviseur))%10 ) ], Tab[j]); // on evalue le i-eme chiffre du nombre et on enfile le nombre dans la file correspondante a ce chiffre.
      // on utilise abs() de math.h car ce sont des nombres negatifs et nous voulons chaque digit en chiffre positif. Sans cela, les chiffres évalues avec cette formule seraient negatifs et on ne pourrait pas utiliser le tableau de files de cette maniere. 
    }
    
    // on reforme le tableau d'origine mais trié jusqu'au i-eme chiffre des nombres en partant du chiffre des unités 
    k =0;
    
    for(j=9; j>=0; j--){ // on defile et range dans le tableau tous les nombres des 10 files dans l'ordre croissant des chiffres (d'abord file 9, 8,7,6 ... car les nombres sont negatifs et -9 < -8 etc ). On utilise une structure FIFO ====> tri stable
      
      while(fileTab[j].head){
	Tab[k] = dequeue(&fileTab[j]);
	k++;
      } 
    }
  }
}






void (*listeFonctionsRdxSort[4])(int*, const int, int) = {rdxSortPosIntTab_Bi,rdxSortPosIntFile_Bi,rdxSortPosIntFile_De,rdxSortNegIntFile_De};



void rdxSortIntChoix(int *Tab, const int SIZE, int TAILLE_CLEF, int rdxSortType){

  int nbPositifs = 0, nbNegatifs = 0, i, n = 0, p = 0;
  int * positivTab = NULL;
  int * negativTab = NULL;

  // calcul de combien de nombres positifs et combien de negatifs dans le tableau 
  for(i=0; i<SIZE; i++){
    if(Tab[i] >= 0){nbPositifs++;}
    else {nbNegatifs++;}
  }

  if(nbPositifs == SIZE){ // s'il n y a que des nombres positifs
    (listeFonctionsRdxSort[rdxSortType])(Tab,SIZE,TAILLE_CLEF); // tri avec le choix demande qui est renseigne via la valeur de l'argument rdxSortType
    return;
  }
  
  else if(nbNegatifs == SIZE){ // s'il n y a que des nombres negatifs 
    /* negativTab = (int*)malloc(nbNegatifs*sizeof(int));
    for(i=0;i<SIZE;i++){
      negativTab[i] = -1*Tab[i];
    }
    
    (listeFonctionsRdxSort[rdxSortType])(negativTab,SIZE,TAILLE_CLEF);

    for(i=nbNegatifs-1; i>= 0; i--){
      Tab[SIZE-1-i] = -1 * negativTab[i];
    }
    
    free(negativTab);*/

    (listeFonctionsRdxSort[3])(Tab,SIZE,TAILLE_CLEF); // tri avec la fonction de tri radix sort pour les tableaux de nombres negatifs 
    return;
  }

  else { // sinon c'est qu'il y a des nombres positifs et des nombres negatifs dans le tableau
    
    positivTab = (int*)malloc(nbPositifs * sizeof(int)); // tableau pour reunir et stocker les nombres positifs
    negativTab = (int*)malloc(nbNegatifs*sizeof(int)); // tableau pour reunir et stocker les nombres negatifs

    // remplissage des tableaux correctement et dans l'ordre d'apparition dans le tableau Tab (tri stable)
    for(i=0;i<SIZE;i++){
      if(Tab[i] < 0){
	negativTab[n] = Tab[i];
	n++;
      }
      else{
	positivTab[p] = Tab[i];
	p++;
      }
    }

    (listeFonctionsRdxSort[rdxSortType])(positivTab,p,TAILLE_CLEF); // tri de la partie positive avec le choix demande qui est renseigne via la valeur de l'argument rdxSortType
    (listeFonctionsRdxSort[3])(negativTab,n,TAILLE_CLEF); // tri de la partie negative avec la fonction de tri radix sort pour les tableaux de nombres negatifs

    // fusion des deux "sous tableaux" tries en commencant par les negatifs pour avoir le tableau d'origine trie entierement
    for(i=0;i<n;i++){
      Tab[i] = negativTab[i];
    }

    for(n=0; n<p; n++){
      Tab[i] = positivTab[n];
      i++;
    }
    
    free(positivTab);
    free(negativTab);

    return;
  }
  
}
